import logging
from datetime import datetime
from pathlib import Path
from typing import List, Union

from logninja.options import All, Only

LOG_RECORD_BUILTIN_ATTRS = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


def get_message(record: logging.LogRecord, max_len: int | None) -> str:
    message = record.getMessage()
    import json

    try:
        result = json.loads(message)
        if result.get("message"):
            message = result.get("message")
    except Exception:
        pass

    if max_len is None:
        return message

    message_size = len(message)
    if message_size == max_len:
        return message

    if message_size > max_len:
        message = message[:max_len] + "..."
        return message

    if message_size < max_len:
        message = message + " " * (max_len + 3 - message_size)

    return message


def format_path(record: logging.LogRecord) -> str:
    path = Path(record.pathname).name
    result = f"{path}:{record.lineno}"
    return result


def get_all_extras(record: logging.LogRecord) -> dict:
    extras_result = {}
    for key, val in record.__dict__.items():
        if key not in LOG_RECORD_BUILTIN_ATTRS:
            extras_result[key] = val
    return extras_result


def get_only_extras(record: logging.LogRecord, extras: List[str]) -> dict:
    extras_result = {}
    for key in extras:
        if key in record.__dict__:
            extras_result[key] = record.__dict__[key]
    return extras_result


def get_extras(
    record: logging.LogRecord,
    extras: Union[All, Only, None],
) -> dict:
    if extras is None:
        return {}

    if isinstance(extras, All):
        return get_all_extras(record)
    if isinstance(extras, Only):
        return get_only_extras(record, extras.extras)

    raise Exception("Invalid option, use All(),  Only([list of extras]) or None")


def get_level_name(record: logging.LogRecord) -> str:
    levelno = record.levelno
    levels = {
        logging.CRITICAL: "[CRITICAL]",
        logging.ERROR: "[ERROR   ]",
        logging.WARNING: "[WARNING ]",
        logging.INFO: "[INFO    ]",
        logging.DEBUG: "[DEBUG   ]",
    }
    level_name = levels.get(levelno)
    return level_name


def format_time(record: logging.LogRecord, fmt: str = "%Y-%m-%d %H:%M:%S") -> str:
    timestamp = record.created
    dt_object = datetime.fromtimestamp(timestamp)
    formatted_time = dt_object.strftime(fmt)
    return f"[{formatted_time}]"


def format_extras(extras: dict) -> str:
    text = ""
    if extras:
        last_key = list(extras.keys())[-1]
        for key, value in extras.items():
            key_value = f"{key}={value}"
            text += key_value
            if key != last_key:
                text += ", "
        return f"  [{text}]"

    return text



import logging
from dataclasses import dataclass, field
from typing import List

from logninja.console_interface import ConsoleInterface
from logninja.ninja_console import NinjaConsole
from logninja.ninja_json_formatter import NinjaJsonFormatter


@dataclass
class LogFileConfig:
    level: int = logging.INFO
    fmt: logging.Formatter = NinjaJsonFormatter()
    filename: str = "logs.jsonl"
    clear_file_on_setup: bool = False


@dataclass
class LogConsoleConfig:
    level: int = logging.INFO
    fmt: logging.Formatter = NinjaJsonFormatter()
    console: ConsoleInterface = NinjaConsole()
    print_exeption: bool = True
    sys_excepthook: bool = False


@dataclass
class RootLoggerConfig:
    level: int = logging.DEBUG
    filters: List[logging.Filter] = field(default_factory=lambda: [])



from abc import ABC, abstractmethod


class ConsoleInterface(ABC):
    """Abstract base class for console interfaces."""

    @abstractmethod
    def print(self, *objects: str, **kwargs):
        """Prints the given message to the console."""
        pass

    @abstractmethod
    def print_json(self, json: str, *args, **kwargs):
        """Prints the given message as JSON to the console."""
        pass

    @abstractmethod
    def print_exception(self, show_locals: bool = False):
        """Prints the exception to the console."""
        pass



import re
from typing import List

try:
    from rich.console import Console
except ImportError:
    raise ImportError(
        "rich library is required for NinjaRichConsole, install it with pip install rich"
    )

HTTP_METHODS: List[str] = [
    "GET",
    "POST",
    "HEAD",
    "PUT",
    "DELETE",
    "OPTIONS",
    "TRACE",
    "PATCH",
]


class NinjaRichConsole(Console):
    def print(self, *objects: str, **kwargs) -> None:
        if all(isinstance(i, str) for i in objects):
            objects = [self._custom_theme(i) for i in objects]
            message = "".join(objects)
            super().print(message, **kwargs)
            return
        super().print(*objects, **kwargs)

    def _custom_theme(self, message: str) -> str:
        message = self._replace_loglevel(message)
        message = self._replace_datetime(message)
        message = self._replace_file_line(message)
        message = self._replace_parentheses_content(message)
        message = self._replace_extras(message)
        message = self._highlight_msg(message)
        return message

    def _replace_loglevel(self, message: str) -> str:
        message = message.replace("[DEBUG   ]", "[blue][DEBUG   ][/blue]")
        message = message.replace("[INFO    ]", "[green][INFO    ][/green]")
        message = message.replace("[WARNING ]", "[yellow][WARNING ][/yellow]")
        message = message.replace("[ERROR   ]", "[bold red][ERROR   ][/bold red]")
        message = message.replace("[CRITICAL]", "[bold red][CRITICAL][/bold red]")
        return message

    def _replace_datetime(self, message: str) -> str:
        string = re.sub(
            r"(\[\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\])", r"[dim]\g<0>[/dim]", message
        )
        return string

    def _replace_file_line(self, message: str):
        string = re.sub(
            r"(\w+\.py:\d+)", r"[dim underline]\g<0>[/dim underline]", message
        )
        return string

    def _replace_parentheses_content(self, message: str) -> str:
        message = re.sub(r"(\(\w+\))", r"[orange4]\g<0>[/orange4]", message)
        return message

    def _replace_extras(self, message: str) -> str:
        result = re.sub(r"(?<=\=)([\w-]+)", r"[blue]\g<0>[/blue]", message)
        result = re.sub(r"([\w-]+)(?=\=)", r"[yellow]\g<0>[/yellow]", result)
        return result

    def _highlight_msg(self, message: str) -> str:
        message = self._highlight_http_status_codes(message)
        message = self._highlight_http_methods(message)
        return message

    def _highlight_http_methods(self, message: str) -> str:
        for method in HTTP_METHODS:
            if method in message:
                message = message.replace(method, f"[bold]{method}[/bold]")
        return message

    def _highlight_http_status_codes(self, message: str) -> str:
        status_codes = {
            "1..": "blue",  # Informational
            "2..": "green",  # Success
            "3..": "yellow",  # Redirection
            "4..": "red",  # Client errors
            "5..": "magenta",  # Server errors
        }

        for status_code, color in status_codes.items():
            status_code_pattern = status_code.replace("..", "\d\d")
            pattern = re.compile(rf"\b{status_code_pattern}\b")
            matches = pattern.findall(message)
            for match in matches:
                message = message.replace(
                    match, f"[bold {color}]{match}[/bold {color}]"
                )
        return message



import logging

from logninja.configs import LogConsoleConfig, LogFileConfig, RootLoggerConfig
from logninja.logger import logger
from logninja.ninja_handler import NinjaHandler


def setup_logging(
    log_file_config: LogFileConfig = None,
    log_console_config: LogConsoleConfig = LogConsoleConfig(),
    root_logger_config: RootLoggerConfig = RootLoggerConfig(),
) -> None:
    if log_file_config is None and log_console_config is None:
        raise ValueError(
            "At least one of log_file_config or log_console_config must be provided"
        )

    root_logger = logging.getLogger()
    root_logger.setLevel(root_logger_config.level)

    if log_console_config:
        console_handler = _setup_log_console_handler(log_console_config)
        root_logger.addHandler(console_handler)
        logger.debug("Console logging setup complete")

    if log_file_config:
        file_handler = _setup_log_file_handler(log_file_config)
        root_logger.addHandler(file_handler)
        logger.debug("File logging setup complete")

    for handler in root_logger.handlers:
        for filter_ in root_logger_config.filters:
            handler.addFilter(filter_)

    logger.debug("Logging setup complete", extra=dict(users="adfdfs"))


def _setup_log_console_handler(
    log_console_config: LogConsoleConfig,
) -> logging.StreamHandler:
    console_handler = NinjaHandler(
        formatter=log_console_config.fmt,
        console=log_console_config.console,
        print_exception=log_console_config.print_exeption,
        sys_excepthook=log_console_config.sys_excepthook,
    )
    console_handler.setLevel(log_console_config.level)
    return console_handler


def _setup_log_file_handler(log_file_config: LogFileConfig) -> logging.FileHandler:
    if log_file_config.clear_file_on_setup:
        with open(log_file_config.filename, "w") as f:
            f.write("")

    file_handler = logging.FileHandler(log_file_config.filename)
    file_handler.setFormatter(fmt=log_file_config.fmt)
    file_handler.setLevel(log_file_config.level)

    return file_handler



import copy
import json
import logging
import sys
from typing import Union

from logninja.console_interface import ConsoleInterface
from logninja.ninja_console import NinjaConsole
from logninja.ninja_formatter import NinjaFormatter
from logninja.ninja_json_formatter import NinjaJsonFormatter


class NinjaHandler(logging.StreamHandler):
    def __init__(
        self,
        formatter: Union[NinjaJsonFormatter, NinjaFormatter],
        console: ConsoleInterface = NinjaConsole(),
        print_exception: bool = True,
        sys_excepthook: bool = False,
    ):
        super().__init__()
        if not isinstance(formatter, (NinjaJsonFormatter, NinjaFormatter)):
            raise ValueError("Invalid formatter")

        self.sys_excepthook = sys_excepthook
        self.print_exception = print_exception
        self.console = console
        self.setFormatter(formatter)

    def emit(self, record: logging.LogRecord) -> None:
        formatted_message = self.format(record)
        if self.is_json(formatted_message):
            self.console.print_json(formatted_message)
        else:
            self.console.print(formatted_message)
        self.flush()

        if record.exc_info and self.print_exception:
            self.console.print_exception(show_locals=True)

        def handle_exception(exc_type, exc_value, exc_traceback):
            if issubclass(exc_type, KeyboardInterrupt):
                sys.__excepthook__(exc_type, exc_value, exc_traceback)
                return

        if self.sys_excepthook is False:
            sys.excepthook = handle_exception

    def is_json(self, message: str) -> bool:
        try:
            message_copy = copy.deepcopy(message)
            json.loads(message_copy)
            return True
        except Exception:
            return False



import datetime as dt
import json
import logging
from typing import Union

from logninja.options import All, Only
from logninja.traceback import format_exception
from logninja.utils import get_extras

FMT_KEYS = {
    "level": "levelname",
    "message": "message",
    "timestamp": "timestamp",
    "logger": "name",
    "module": "module",
    "function": "funcName",
    "line": "lineno",
    "thread_name": "threadName",
}


class NinjaJsonFormatter(logging.Formatter):
    def __init__(self, extras: Union[All, Only, None] = All):
        super().__init__()
        self.extras = extras

    def format(self, record: logging.LogRecord) -> str:
        message = self._prepare_log_dict(record)
        return json.dumps(message, default=str)

    def _prepare_log_dict(self, record: logging.LogRecord):
        always_fields = {
            "message": record.getMessage(),
            "timestamp": dt.datetime.fromtimestamp(
                record.created, tz=dt.timezone.utc
            ).isoformat(),
        }
        if record.exc_info is not None:
            always_fields["exc_info"] = format_exception(exc_info=record.exc_info)

        message = {
            key: (
                msg_val
                if (msg_val := always_fields.pop(val, None)) is not None
                else getattr(record, val)
            )
            for key, val in FMT_KEYS.items()
        }
        message.update(always_fields)

        extras = get_extras(record=record, extras=self.extras)
        for key, value in extras.items():
            message[key] = value

        return message



from typing import List


class All:
    def __repr__(self):
        return "All()"


class Only:
    def __init__(self, keys: List[str]):
        if not isinstance(keys, list):
            raise ValueError("keys must be a list")
        self.extras = keys

    def __repr__(self):
        return f"Only({self.extras})"



import json
import sys

from logninja.console_interface import ConsoleInterface
from logninja.traceback import format_exception


class NinjaConsole(ConsoleInterface):
    terminator = "\n"

    def __init__(self) -> None:
        self.stream = sys.stderr

    def print(self, *objects: str, **kwargs) -> None:
        try:
            stream = self.stream
            final_message = "".join(objects)
            stream.write(final_message + self.terminator)
        except RecursionError:
            raise

    def print_json(self, json: str, *args, **kwargs) -> None:
        self.print(json, *args, **kwargs)

    def print_exception(self, show_locals: bool = False):
        exc_info = sys.exc_info()
        trace = format_exception(exc_info)
        print(json.dumps(trace, indent=4))



import logging


class FormatFilter(logging.Filter):
    """
    In some cases, we need to format the log at the filter level.
    This class provides a way to do that. It takes a formatter and applies it
    to the log. Here's an example use case:
        - When using the instrument-opentelemetry for logging,the formatter is
        never calledat the handler level.Therefore, we need to format the log at
        the filter level.
    """

    def __init__(
        self,
        formatter: logging.Formatter,
        name: str = "FormatFilter",
    ) -> None:
        super().__init__(name)
        self.formatter = formatter

    def filter(self, record: logging.LogRecord):
        record.msg = self.formatter.format(record)
        return True



import asyncio
import functools
import logging
import time

from logninja.logger import logger as logninja_logger


def log_execution(
    logger: logging.Logger = logninja_logger,
    level: int = logging.INFO,
    capture_exception: bool = False,
):
    """
    Decorator to log the execution time of a function or coroutine.

    Args:
        logger (logging.Logger): The logger to be used.
        level (int): The logging level to be used.
        capture_exception (bool): Whether to capture exceptions. Defaults to False.

    Returns:
        Callable: The decorator.
    """

    def decorator(func):
        levels = {
            logging.CRITICAL: "critical",
            logging.ERROR: "error",
            logging.WARNING: "warning",
            logging.INFO: "info",
            logging.DEBUG: "debug",
        }
        level_attr = levels.get(level, "info")
        logger_with_level = getattr(logger, level_attr)

        def log_start(*args, **kwargs):
            params = {"args": list(args), "kwargs": kwargs}
            logger_with_level(
                f"Starting execution of function '{func.__qualname__}'",
                extra={"params": str(params)},
            )

        def log_end(end: float, start: float):
            logger_with_level(
                f"'{func.__qualname__}' executed in {end - start:.4f} seconds"
            )

        def log_exception(exc: Exception):
            logger.exception(
                f"An error occurred while executing '{func.__qualname__}'",
                exc_info=exc,
            )
            raise exc

        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            log_start(*args, **kwargs)

            start = time.time()
            if capture_exception:
                try:
                    result = func(*args, **kwargs)
                except Exception as exc:
                    log_exception(exc)
                    return exc
            else:
                result = func(*args, **kwargs)
            end = time.time()

            log_end(end, start)
            return result

        async def async_wrapper(*args, **kwargs):
            log_start(*args, **kwargs)

            start = time.time()
            if capture_exception:
                try:
                    result = await func(*args, **kwargs)
                except Exception as exc:
                    log_exception(exc)
                    return exc
            else:
                result = await func(*args, **kwargs)
            end = time.time()

            log_end(end, start)
            return result

        return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper

    return decorator



import logging
from typing import Union

from logninja.options import All, Only
from logninja.utils import (
    format_extras,
    format_path,
    format_time,
    get_extras,
    get_level_name,
    get_message,
)


class NinjaFormatter(logging.Formatter):
    def __init__(
        self, extras: Union[All, Only, None] = None, max_message_length: int | None = 60
    ):
        super().__init__()
        self.extras = extras
        self.max_message_length = max_message_length

    def format(self, record: logging.LogRecord) -> str:
        message = get_message(record, self.max_message_length)

        message_time = format_time(record)
        message_level = get_level_name(record)
        filepath = format_path(record)
        extras = get_extras(record=record, extras=self.extras)
        formatted_extras = format_extras(extras)

        structured_message = (
            f"{message_time}{message_level} {message} ({record.name}) {filepath}"
            + formatted_extras
        )

        return structured_message
